from enum import Enum


class TokenType(Enum):
    INT = 0,
    FLOAT = 1,
    DOUBLE = 2,
    LONG = 3,
    HEX = 4,
    BIN = 5,
    IDENTIFIER = 6,
    HARD_KEYWORD = 7,
    SOFT_KEYWORD = 8,
    MODIFIER_KEYWORD = 9,
    OPERATOR = 10,
    SEPARATOR = 11,
    DOT = 12,
    BRACKET = 13,
    CHAR = 14,
    STRING = 15,
    COMMENT = 16,
    MULTILINE_COMMENT = 17,
    WHITESPACE = 18,
    NEW_LINE = 19,
    SPECIAL_SYMBOL = 20,
    ERROR = 21,
    DOC_COMMENT = 22


class Token:
    def __init__(self, token_type, end, value=None):
        self.token_type = token_type
        self.end = end
        self.value = value

    def to_string(self):
        if self.value is not None:
            return "{" + self.token_type.name + " | " + self.value + "}"




HARD_KEYWORDS = (
    "as", "break", "catch", "class", "continue", "do", "else", "false", "for", "fun", "if", "in", "interface", "is",
    "null", "object", "package", "return", "super", "this", "throw", "true", "try", "typealias", "typeof",
    "val", "var", "when", "while")

SOFT_KEYWORDS = (
    "by", "constructor", "delegate", "dynamic", "field", "file", "finally", "get", "import", "init",
    "it", "param", "property", "receiver", "set", "setparam", "where")

MODIFIER_KEYWORDS = (
    "actual", "abstract", "annotation", "companion", "const", "crossinline", "data", "enum", "expect",
    "external", "final", "infix", "inline", "inner", "internal", "lateinit", "noinline", "open", "operator",
    "out", "override", "private", "protected", "public", "reified", "sealed", "suspend", "tailrec", "vararg")



import argparse
from .ccf import *


def main():
    # print(bool(re.match("^[a-z][A-Za-z]*$", '_camelCase')))
    parser = argparse.ArgumentParser('KotlinCCF')
    parser.add_argument('-v', '--verify', action='store_true', help='Verify naming conventions')
    parser.add_argument('-fx', '--fix', action='store_true', help='Fix naming')
    parser.add_argument('-f', '--file', nargs=1, required=False, help='*.kt file')
    parser.add_argument('-p', '--project', nargs=1, required=False, help='Project with *.kt files')
    parser.add_argument('-d', '--directory', nargs=1, required=False, help='Directory with *.kt files')

    my_namespace = parser.parse_args()
    if not (my_namespace.verify or my_namespace.fix) or \
            not (my_namespace.directory is not None or
                 my_namespace.file is not None or
                 my_namespace.project is not None):
        print("error")
        parser.print_help()
        return

    ccf = CodeConventionsFixer()
    if my_namespace.fix:
        ccf.set_type(CCFType.FIX)
    else:
        ccf.set_type(CCFType.VERIFY)
    if my_namespace.directory is not None:
        ccf.set_file_walk_type(FileWalkType.DIRECTORY)
        ccf.run(my_namespace.directory[0])
    elif my_namespace.project is not None:
        ccf.set_file_walk_type(FileWalkType.PROJECT)
        ccf.run(my_namespace.project[0])
    else:
        ccf.run(my_namespace.file[0])




main()



from .lexer import *
from enum import Enum
import os
import logging
import re


class CCFType(Enum):
    FIX = 0,
    VERIFY = 1


class FileWalkType(Enum):
    FILE = 0,
    PROJECT = 1,
    DIRECTORY = 2


class CodeConventionsFixer:
    def __init__(self):
        self.verify_log_file_name = "verification.log"
        self.fix_log_file_name = "fixing.log"
        self.file_walk_type = FileWalkType.FILE
        self.ccf_type = CCFType.VERIFY
        self.log_file = self.verify_log_file_name
        self.reformat_classes_names = {}
        self.reformat_classes_packages = {}
        self.reformat_classes_packages_new_names = {}
        self.renamed_packages = {}

    def set_file_walk_type(self, file_walk_type):
        self.file_walk_type = file_walk_type

    def set_type(self, ccf_type):
        self.ccf_type = ccf_type
        if ccf_type == CCFType.FIX:
            self.log_file = self.fix_log_file_name

        logger = logging.getLogger()
        fhandler = logging.FileHandler(filename=self.log_file, mode='w')
        formatter = logging.Formatter('%(levelname)s: %(message)s')
        fhandler.setFormatter(formatter)
        logger.addHandler(fhandler)
        logger.setLevel(logging.DEBUG)

    def to_string(self):
        return self.file_walk_type.name + " " + self.ccf_type.name

    def run(self, path):
        if self.ccf_type == CCFType.VERIFY:
            logging.basicConfig(filename=self.verify_log_file_name)
        else:
            logging.basicConfig(filename=self.fix_log_file_name)

        if self.file_walk_type == FileWalkType.FILE:
            self.format_file(path)
        elif self.file_walk_type == FileWalkType.PROJECT:
            self.format_project(path)
        else:
            self.format_dir(path)

    def format_tokens(self, path):
        # class/object/interface PascalCase
        # val/fun camelCase
        # const val MAX_COUNT = 8
        # val USER_NAME_FIELD = "UserName"
        # fun processDeclarations(){ / * ... * /}
        # var declarationCount = 1

        output = ""
        lexer = Lexer()
        indent_level = 0
        line = 1
        last_bracket_index = 0
        package_name = ""
        new_package_name = ""
        is_modified = False
        refactor_names = {}
        refactor_indents = {}
        val_types = [TokenType.INT, TokenType.DOUBLE, TokenType.FLOAT, TokenType.LONG, TokenType.STRING, TokenType.CHAR]
        with open(path, 'r', encoding="utf-8") as file:
            orig_tokens = lexer.tokenize(file.read())
            self.refactor_docs(orig_tokens)
            tokens = list(filter(lambda x: x.token_type != TokenType.WHITESPACE, orig_tokens))
            for i in range(0, len(tokens)):
                token = tokens[i]
                if token.value == "{":
                    indent_level += 1
                    last_bracket_index = i
                elif token.value == "}":
                    indent_level -= 1
                    names_to_del = list()
                    for name in refactor_indents.keys():
                        level = refactor_indents[name]
                        if level is not None and indent_level < level:
                            names_to_del.append(name)
                    for name in names_to_del:
                        del refactor_indents[name]
                        del refactor_names[name]
                elif token.token_type == TokenType.HARD_KEYWORD and token.value == "package":
                    for j in range(i + 1, len(tokens)):
                        if tokens[j].token_type == TokenType.DOT or tokens[j].token_type == TokenType.IDENTIFIER:
                            package_name += tokens[j].value
                            if tokens[j].token_type == TokenType.IDENTIFIER and not self.is_camel_case(tokens[j].value):
                                tokens[j].value = self.to_camel_case(tokens[j].value)
                                is_modified = True
                            new_package_name += tokens[j].value
                        else:
                            break
                    if is_modified:
                        self.write_log(path, line, "PACKAGE_NAMING_ERROR", package_name, new_package_name)
                        self.reformat_classes_packages_new_names[package_name] = new_package_name
                elif token.token_type == TokenType.HARD_KEYWORD:
                    if token.value == "val":
                        if tokens[i + 1].token_type == TokenType.IDENTIFIER and tokens[i + 2].value == "=" and \
                                tokens[i + 3].token_type in val_types:
                            if not self.is_upper_snake_case(tokens[i + 1].value):
                                refactor_names[tokens[i + 1].value] = self.to_upper_snake_case(tokens[i + 1].value)
                                refactor_indents[tokens[i + 1].value] = indent_level
                                self.refactor_prev(tokens[last_bracket_index + 1: i], tokens[i + 1].value,
                                                   self.to_upper_snake_case(tokens[i + 1].value))
                                self.write_log(path, line, "VAL_NAMING_ERROR", tokens[i + 1].value,
                                               self.to_upper_snake_case(tokens[i + 1].value))
                        else:
                            if not self.is_camel_case(tokens[i + 1].value):
                                refactor_names[tokens[i + 1].value] = self.to_camel_case(tokens[i + 1].value)
                                refactor_indents[tokens[i + 1].value] = indent_level
                                self.refactor_prev(tokens[last_bracket_index + 1: i], tokens[i + 1].value,
                                                   self.to_camel_case(tokens[i + 1].value))
                                self.write_log(path, line, "VAL_NAMING_ERROR", tokens[i + 1].value,
                                               self.to_camel_case(tokens[i + 1].value))
                    elif token.value == "var":
                        if tokens[i + 1].token_type == TokenType.IDENTIFIER and tokens[i + 2].value == "=":
                            if not self.is_camel_case(tokens[i + 1].value):
                                refactor_names[tokens[i + 1].value] = self.to_camel_case(tokens[i + 1].value)
                                refactor_indents[tokens[i + 1].value] = indent_level
                                self.refactor_prev(tokens[last_bracket_index + 1: i], tokens[i + 1].value,
                                                   self.to_camel_case(tokens[i + 1].value))
                                self.write_log(path, line, "VAR_NAMING_ERROR", tokens[i + 1].value,
                                               self.to_camel_case(tokens[i + 1].value))
                    elif token.value == "fun":
                        if tokens[i + 1].token_type == TokenType.IDENTIFIER and tokens[i + 2].value == "(" and \
                                tokens[i - 1].value != "Test":
                            if tokens[i + 3].value == ")" and tokens[i + 1].value == tokens[i + 5].value:
                                pass
                            elif not self.is_camel_case(tokens[i + 1].value):
                                refactor_names[tokens[i + 1].value] = self.to_camel_case(tokens[i + 1].value)
                                refactor_indents[tokens[i + 1].value] = indent_level
                                self.refactor_prev(tokens[last_bracket_index + 1: i], tokens[i + 1].value,
                                                   self.to_camel_case(tokens[i + 1].value))
                                self.write_log(path, line, "FUN_NAMING_ERROR", tokens[i + 1].value,
                                               self.to_camel_case(tokens[i + 1].value))
                    elif token.value == "object":
                        if tokens[i + 1].token_type == TokenType.IDENTIFIER:
                            if not self.is_pascal_case(tokens[i + 1].value):
                                refactor_names[tokens[i + 1].value] = self.to_pascal_case(tokens[i + 1].value)
                                refactor_indents[tokens[i + 1].value] = indent_level
                                self.refactor_prev(tokens[last_bracket_index + 1: i], tokens[i + 1].value,
                                                   self.to_pascal_case(tokens[i + 1].value))
                                self.write_log(path, line, "OBJECT_NAMING_ERROR", tokens[i + 1].value,
                                               self.to_camel_case(tokens[i + 1].value))
                    elif token.value == "class" or token.value == "interface":
                        if not self.is_pascal_case(tokens[i + 1].value):
                            refactor_names[tokens[i + 1].value] = self.to_pascal_case(tokens[i + 1].value)
                            refactor_indents[tokens[i + 1].value] = indent_level
                            self.refactor_prev(tokens[last_bracket_index + 1: i], tokens[i + 1].value,
                                               self.to_pascal_case(tokens[i + 1].value))
                            self.reformat_classes_names[tokens[i + 1].value] = self.to_pascal_case(tokens[i + 1].value)
                            self.reformat_classes_packages[tokens[i + 1].value] = package_name
                            self.write_log(path, line, token.value.upper() + "_NAMING_ERROR", tokens[i + 1].value,
                                           self.to_pascal_case(tokens[i + 1].value))

                elif token.token_type == TokenType.IDENTIFIER:
                    if token.value in refactor_names.keys() is not None:
                        token.value = refactor_names[token.value]
                elif token.token_type == TokenType.NEW_LINE:
                    line += 1
        k = 0
        for t in orig_tokens:
            if t.token_type == TokenType.WHITESPACE:
                output += t.value
            elif k < len(tokens):
                output += tokens[k].value
                k += 1
        return output

    def refactor_docs(self, orig_tokens):
        tokens = list(filter(lambda x: x.token_type == TokenType.DOC_COMMENT, orig_tokens))
        for token in tokens:
            i = orig_tokens.index(token)
            indent = orig_tokens[i - 1].value if i > 0 and orig_tokens[i - 1].token_type == TokenType.WHITESPACE else ""
            lines = token.value.split("\n")
            if len(lines) != 1:
                if lines[0] != "/**":
                    token.value = lines[0][0:3] + "\n" + indent + " * " + lines[0][3:len(lines[0])].lstrip() + "\n"
                else:
                    token.value = lines[0] + "\n"
                for line in lines[1:len(lines) - 1]:
                    token.value += indent + " * " + line.lstrip(" *") + "\n"
                if lines[len(lines) - 1].lstrip() != "*/":
                    # lines[len(lines) - 1]
                    token.value += indent + " * " + lines[len(lines) - 1].lstrip()[
                                                    0:len(lines) - 2] + "\n" + indent + " */"
                else:
                    token.value += indent + " */"

    def write_log(self, file_path, line_number, error_type, error_val, correct_val):
        # Id. File Path: Line Number - Error Code: Error Message
        if self.ccf_type == CCFType.FIX:
            logging.error(file_path + ": " + str(line_number) + "-" + error_type + ": "
                          + error_val + " renamed to " + correct_val)
        else:
            logging.warning(file_path + ": " + str(line_number) + "-" + error_type + ": "
                            + error_val + " should be " + correct_val)

    def refactor_prev(self, tokens, old, new):
        for token in tokens:
            if token.token_type == TokenType.IDENTIFIER:
                if token.value == old:
                    token.value = new

    def refactor_classes(self, files):
        lexer = Lexer()
        for f in files:
            output = ""
            reformat_classes = {}
            with open(f, 'r', encoding="utf-8") as file:
                orig_tokens = lexer.tokenize(file.read())
                tokens = list(
                    filter(lambda x: x.token_type != TokenType.WHITESPACE and x.token_type != TokenType.NEW_LINE,
                           orig_tokens))
                for i in range(0, len(tokens)):
                    token = tokens[i]
                    if token.token_type == TokenType.SOFT_KEYWORD and token.value == "import":
                        imp = ""
                        for j in range(i + 1, len(tokens)):
                            if tokens[j].token_type == TokenType.DOT or \
                                    tokens[j].token_type == TokenType.IDENTIFIER or tokens[j].value == "*":
                                imp += tokens[j].value
                            else:
                                break
                        for p in self.reformat_classes_packages.keys():
                            if self.is_imported(imp, self.reformat_classes_packages[p], p):
                                reformat_classes[p] = self.reformat_classes_names[p]
                                for j in range(i + 1, len(tokens)):
                                    if tokens[j].token_type == TokenType.DOT or \
                                            tokens[j].token_type == TokenType.IDENTIFIER or tokens[j].value == "*":
                                        if tokens[j].token_type == TokenType.IDENTIFIER:
                                            tokens[j].value = self.to_camel_case(tokens[j].value)
                                    else:
                                        break
                                break
                        for p in self.reformat_classes_packages_new_names.keys():
                            if self.is_imported(imp, p, ""):
                                for j in range(i + 1, len(tokens)):
                                    if tokens[j].token_type == TokenType.DOT or \
                                            tokens[j].token_type == TokenType.IDENTIFIER or tokens[j].value == "*":
                                        if tokens[j].token_type == TokenType.IDENTIFIER:
                                            tokens[j].value = self.to_camel_case(tokens[j].value)
                                    else:
                                        break
                                break
                    elif token.token_type == TokenType.IDENTIFIER:
                        if token.value in reformat_classes.keys() is not None:
                            token.value = reformat_classes[token.value]

            k = 0
            for t in orig_tokens:
                if t.token_type in (TokenType.NEW_LINE, TokenType.WHITESPACE):
                    output += t.value
                elif k < len(tokens):
                    output += tokens[k].value
                    k += 1
            if self.ccf_type == CCFType.FIX:
                self.save_formatted_file(output, f)

    def is_imported(self, imp, package, class_name):
        i = 0
        while True:
            if i == len(imp) or i == len(package) or package[i] != imp[i]:
                break
            else:
                i += 1
        if i == len(package) and i < len(imp) and class_name != "" and imp[i: len(imp)] == "." + class_name:
            return True
        elif i < len(imp) and ((imp[i - 1] == "." and imp[i] == "*") or (i + 1 < len(imp) and imp[i] == "." and imp[i + 1] == "*")):
            return True
        return False

    def format_project(self, path):
        files = []
        if self.ccf_type == CCFType.FIX:
            while True:
                tree = os.walk(path)
                for d in tree:
                    cur_dir_name = d[0]
                    dir_path = cur_dir_name.split("/")
                    if not self.is_camel_case(dir_path[len(dir_path) - 1]):
                        self.renamed_packages[dir_path[len(dir_path) - 1]] = \
                            self.to_camel_case(dir_path[len(dir_path) - 1])
                        dir_path[len(dir_path) - 1] = self.to_camel_case(dir_path[len(dir_path) - 1])
                        os.rename(cur_dir_name, "/".join(dir_path))
                        continue
                break

        tree = os.walk(path)
        for d in tree:
            cur_dir_name = d[0]
            cur_dir_files = d[2]
            for file in cur_dir_files:
                if file.endswith(".kt"):
                    if not self.is_pascal_case(file[0: len(file) - 3]):
                        if self.ccf_type == CCFType.FIX:
                            os.rename(cur_dir_name + '/' + file, cur_dir_name + '/'
                                      + self.to_pascal_case(file[0: len(file) - 3]) + ".kt")
                        self.write_log(cur_dir_name + '/' + file, 0, "FILE_NAMING_ERROR",
                                       file, self.to_pascal_case(file[0: len(file) - 3]) + ".kt")
                        if self.ccf_type == CCFType.FIX:
                            files.append(cur_dir_name + '/' + self.to_pascal_case(file[0: len(file) - 3]) + ".kt")
                        else:
                            files.append(cur_dir_name + '/' + file)
                    else:
                        files.append(cur_dir_name + '/' + file)

        for file in files:
            formatted_code = self.format_tokens(file)
            if self.ccf_type == CCFType.FIX:
                self.save_formatted_file(formatted_code, file)
        self.refactor_classes(files)

    def format_dir(self, path):
        files = []
        tree = os.walk(path)
        for d in tree:
            cur_dir_name = d[0]
            cur_dir_files = d[2]
            for file in cur_dir_files:
                if file.endswith(".kt"):
                    if not self.is_pascal_case(file[0: len(file) - 3]):
                        if self.ccf_type == CCFType.FIX:
                            os.rename(cur_dir_name + '/' + file, cur_dir_name + '/'
                                      + self.to_pascal_case(file[0: len(file) - 3]) + ".kt")
                        self.write_log(cur_dir_name + '/' + file, 0, "FILE_NAMING_ERROR",
                                       file, self.to_pascal_case(file[0: len(file) - 3]) + ".kt")
                        if self.ccf_type == CCFType.FIX:
                            files.append(cur_dir_name + '/' + self.to_pascal_case(file[0: len(file) - 3]) + ".kt")
                        else:
                            files.append(cur_dir_name + '/' + file)
                    else:
                        files.append(cur_dir_name + '/' + file)

            break
        for file in files:
            formatted_code = self.format_tokens(file)
            if self.ccf_type == CCFType.FIX:
                self.save_formatted_file(formatted_code, file)
        self.refactor_classes(files)

    def format_file(self, path):
        formatted_code = self.format_tokens(path)
        if self.ccf_type == CCFType.FIX:
            self.save_formatted_file(formatted_code, path)

    def save_formatted_file(self, formatted_code, path):
        with open(path, 'w') as file:
            file.write(formatted_code)

    def is_upper_snake_case(self, value: str):
        result = re.match(r'[A-Z]+(_(([A-Z]+)|([0-9]+)))*', value)
        return True if result is not None and result.start() == 0 and result.end() == len(value) else False

    def is_camel_case(self, value: str):
        result = re.match(r'[a-z]+((\d)|([A-Z0-9][a-z0-9]+))*([A-Z])?', value)
        return True if result is not None and result.start() == 0 and result.end() == len(value) else False

    def is_pascal_case(self, value: str):
        result = self.to_pascal_case(value) == value
        # re.match(r'([A-Z][a-z0-9]+)((\d)|([A-Z0-9][a-z0-9]+))*([A-Z])?', value)
        return result
        # True if result is not None and result.start() == 0 and result.end() == len(value) else False

    def to_upper_snake_case(self, value: str):
        new_name = ""
        last_case = -1  # symb = 0, digit = 1, lower = 2, upper = 3
        for letter in value:
            if letter.isalpha():
                if letter.isupper():
                    if last_case == -1 or last_case == 3 or last_case == 0:
                        new_name += letter
                    else:
                        new_name += "_" + letter
                    last_case = 3
                elif letter.islower():
                    if last_case == 1:
                        new_name += "_" + letter.upper()
                    else:
                        new_name += letter.upper()
                    last_case = 2
            elif letter.isnumeric():
                if last_case == -1 or last_case == 1 or last_case == 0:
                    new_name += letter
                else:
                    new_name += "_" + letter
                last_case = 1
            else:
                if last_case != -1 and last_case != 0:
                    new_name += "_"
                last_case = 0
        if last_case == 0:
            new_name = new_name[0:len(new_name) - 1]
        return new_name

    def to_camel_case(self, value: str):
        new_name = ""
        last_case = -1  # symb = 0, digit = 1, lower = 2, upper = 3
        for letter in value:
            if letter.isalpha():
                if letter.isupper():
                    if last_case == 3 or last_case == -1:
                        new_name += letter.lower()
                    else:
                        new_name += letter
                    last_case = 3
                elif letter.lower():
                    if last_case == 0 or last_case == 1:
                        new_name += letter.upper()
                    else:
                        new_name += letter
                    last_case = 2
            elif letter.isnumeric():
                new_name += letter
                last_case = 1
            else:
                last_case = 0
        return new_name

    def to_pascal_case(self, value: str):
        new_name = ""
        last_case = -1  # symb = 0, digit = 1, lower = 2, upper = 3
        for letter in value:
            if letter.isalpha():
                if letter.isupper():
                    if last_case == 3:
                        new_name += letter.lower()
                    else:
                        new_name += letter
                    last_case = 3
                elif letter.lower():
                    if last_case == 0 or last_case == 1:
                        new_name += letter.upper()
                        last_case = 2
                    elif last_case == -1:
                        new_name += letter.upper()
                        last_case = 3
                    else:
                        new_name += letter
                        last_case = 2

            elif letter.isnumeric():
                new_name += letter
                last_case = 1
            else:
                last_case = 0
        return new_name



from .utils import *
from .l_token import *


def get_type(text):
    if text in HARD_KEYWORDS:
        return TokenType.HARD_KEYWORD
    elif text in SOFT_KEYWORDS:
        return TokenType.SOFT_KEYWORD
    elif text in MODIFIER_KEYWORDS:
        return TokenType.MODIFIER_KEYWORD
    else:
        return TokenType.IDENTIFIER


def identifier_keyword_dfa(position, text):
    cur_position = position
    state = 0
    while cur_position < len(text):
        cur_char = text[cur_position]
        cur_position += 1
        if state == 0:
            if cur_char == '_' or cur_char.isalpha():
                state = 1
            else:
                return None
        elif state == 1:
            if not (cur_char.isalpha() or cur_char.isnumeric() or cur_char == '_' or cur_char == '$'):
                return Token(get_type(text[position:cur_position - 1]), cur_position - 1,
                             text[position:cur_position - 1])

    if state == 1:
        return Token(get_type(text[position:cur_position]), cur_position, text[position:cur_position])


def multiline_comment_dfa(position, text):
    cur_position = position
    state = 0
    while cur_position < len(text):
        cur_char = text[cur_position]
        cur_position += 1
        if state == 0:
            if cur_char == '/':
                state = 1
            else:
                return None
        elif state == 1:
            if cur_char == '*':
                state = 2
            else:
                return None
        elif state == 2:
            if cur_char == '*':
                state = 3
        elif state == 3:
            if cur_char == '/':
                return Token(TokenType.MULTILINE_COMMENT, cur_position, text[position:cur_position])
            else:
                state = 2
    return None


def doc_comment_dfa(position, text):
    cur_position = position
    state = 0
    while cur_position < len(text):
        cur_char = text[cur_position]
        cur_position += 1
        if state == 0:
            if cur_char == '/':
                state = 1
            else:
                return None
        elif state == 1:
            if cur_char == '*':
                state = 2
            else:
                return None
        elif state == 2:
            if cur_char == '*':
                state = 3
            else:
                return None
        elif state == 3:
            if cur_char == '*':
                state = 4
        elif state == 4:
            if cur_char == '/':
                return Token(TokenType.DOC_COMMENT, cur_position, text[position:cur_position])
            else:
                state = 3
    return None


def single_line_comment_dfa(position, text):
    cur_position = position
    state = 0
    while cur_position < len(text):
        cur_char = text[cur_position]
        cur_position += 1
        if state == 0:
            if cur_char == '/':
                state = 1
            else:
                return None
        elif state == 1:
            if cur_char == '/':
                state = 2
            else:
                return None
        elif state == 2:
            if cur_char == '\n':
                return Token(TokenType.COMMENT, cur_position, text[position:cur_position])

    return Token(TokenType.COMMENT, cur_position, text[position:cur_position]) if state == 2 else None


def string_dfa(position, text):
    cur_position = position
    state = 0
    while cur_position < len(text):
        cur_char = text[cur_position]
        cur_position += 1
        if state == 0:
            if cur_char == '"':
                state = 1
            elif cur_char == "'":
                state = 2
            else:
                return None
        elif state == 1:
            if cur_char == '"':
                return Token(TokenType.STRING, cur_position, text[position:cur_position])
        elif state == 2:
            if cur_char != "'":
                state = 3
            else:
                return None
        elif state == 3:
            if cur_char == "'":
                return Token(TokenType.CHAR, cur_position, text[position:cur_position])
            else:
                return None
    return Token(TokenType.COMMENT, cur_position, text[position:cur_position]) if state == 2 else None


def whitespace_dfa(position, text):
    cur_position = position
    state = 0
    while cur_position < len(text):
        cur_char = text[cur_position]
        cur_position += 1
        if state == 0:
            if " \t\r".find(cur_char) != -1:
                state = 1
            elif cur_char == '\n':
                return Token(TokenType.NEW_LINE, cur_position, text[position:cur_position])
            else:
                return None
        elif state == 1:
            if cur_char == '\n':
                return Token(TokenType.NEW_LINE, cur_position, text[position:cur_position])
            elif " \t\r".find(cur_char) == -1:
                return Token(TokenType.WHITESPACE, cur_position - 1, text[position:cur_position - 1])
    if state == 1:
        return Token(TokenType.WHITESPACE, cur_position, text[position:cur_position])
    return None


def number_dfa(position, text):
    cur_position = position
    state = 0
    result = None
    while cur_position < len(text):
        cur_char = text[cur_position]
        cur_position += 1
        if state == 0 or state == 1:
            if cur_char == '0':
                state = 2
                result = Token(TokenType.INT, cur_position, text[position:cur_position])
            elif cur_char.isnumeric():
                state = 4
                result = Token(TokenType.INT, cur_position, text[position:cur_position])
            elif cur_char == '.':
                state = 6
            else:
                return None
        elif state == 2:
            if cur_char.isnumeric() or cur_char == '_':
                state = 3
            elif cur_char == "L":
                return Token(TokenType.LONG, cur_position, text[position:cur_position])
            elif cur_char == '.':
                state = 6
            elif cur_char == "F" or cur_char == "f":
                return Token(TokenType.FLOAT, cur_position, text[position:cur_position])
            else:
                return result
        elif state == 3:
            if cur_char == "F" or cur_char == "f":
                return Token(TokenType.FLOAT, cur_position, text[position:cur_position])
            elif cur_char.isnumeric() or cur_char == '_':
                pass
            else:
                return result
        elif state == 4:
            if cur_char.isnumeric() or cur_char == '_':
                result = Token(TokenType.INT, cur_position, text[position:cur_position])
            elif cur_char == "L":
                return Token(TokenType.LONG, cur_position, text[position:cur_position])
            elif cur_char == "F" or cur_char == "f":
                return Token(TokenType.FLOAT, cur_position, text[position:cur_position])
            elif cur_char == '.':
                state = 6
            else:
                return result
        elif state == 6:
            if cur_char.isnumeric():
                result = Token(TokenType.DOUBLE, cur_position, text[position:cur_position])
                state = 7
            else:
                return result
        elif state == 7:
            if cur_char.isnumeric() or cur_char == '_':
                result = Token(TokenType.DOUBLE, cur_position, text[position:cur_position])
            elif cur_char == "F" or cur_char == "f":
                return Token(TokenType.FLOAT, cur_position, text[position:cur_position])
            else:
                return result
    return result


def operator_dfa(position, text):
    cur_position = position
    state = 0
    result = None
    while cur_position < len(text):
        cur_char = text[cur_position]
        cur_position += 1
        if state == 0:
            if cur_char == '&':
                state = 1
            elif cur_char == '|':
                state = 2
            elif "<>*/%".find(cur_char) != -1:
                state = 3
                result = Token(TokenType.OPERATOR, cur_position, text[position:cur_position])
            elif cur_char == '!' or cur_char == '=':
                state = 4
                result = Token(TokenType.OPERATOR, cur_position, text[position:cur_position])
            elif cur_char == '+':
                state = 5
                result = Token(TokenType.OPERATOR, cur_position, text[position:cur_position])
            elif cur_char == '-':
                state = 6
                result = Token(TokenType.OPERATOR, cur_position, text[position:cur_position])
            else:
                return None
        elif state == 1:
            if cur_char == '&':
                return Token(TokenType.OPERATOR, cur_position, text[position:cur_position])
            else:
                return None
        elif state == 2:
            if cur_char == '|':
                return Token(TokenType.OPERATOR, cur_position, text[position:cur_position])
            else:
                return None
        elif state == 3:
            if cur_char == '=':
                return Token(TokenType.OPERATOR, cur_position, text[position:cur_position])
            else:
                return result
        elif state == 4:
            if cur_char == '=':
                state = 3
                result = Token(TokenType.OPERATOR, cur_position, text[position:cur_position])
            else:
                return result
        elif state == 5:
            if cur_char == '+' or cur_char == '=':
                return Token(TokenType.OPERATOR, cur_position, text[position:cur_position])
            else:
                return result
        elif state == 6:
            if cur_char == '-' or cur_char == '=':
                result = Token(TokenType.OPERATOR, cur_position, text[position:cur_position])
            else:
                return result
        else:
            return result
    return result


def punctuation_dfa(position, text):
    cur_char = text[position]

    if "{}[]()".find(cur_char) != -1:
        return Token(TokenType.BRACKET, position + 1, text[position:position + 1])
    if ",;".find(cur_char) != -1:
        return Token(TokenType.SEPARATOR, position + 1, text[position:position + 1])
    if cur_char == '.':
        return Token(TokenType.DOT, position + 1, text[position:position + 1])
    return None


def special_symbol_dfa(position, text):
    special_symbols = ("?.", "?:", "..", ":", "::", "->", "!!", "$", "@")
    result = None
    if text[position] in special_symbols:
        result = Token(TokenType.SPECIAL_SYMBOL, position + 1, text[position])
    if position + 1 < len(text) and text[position:position + 2] in special_symbols:
        result = Token(TokenType.SPECIAL_SYMBOL, position + 2, text[position:position + 2])
    return result



from .dfaMatchers import *
from .l_token import *

MATCHERS = [
    operator_dfa,
    number_dfa,
    special_symbol_dfa,
    punctuation_dfa,
    identifier_keyword_dfa,
    whitespace_dfa,
    string_dfa,
    single_line_comment_dfa,
    doc_comment_dfa,
    multiline_comment_dfa
]


class Lexer:
    def __init__(self):
        self.tokens = []

    def tokenize(self, code):
        self.tokens.clear()
        position = 0
        error = None
        while position < len(code):
            to_add = None

            for i in range(len(MATCHERS)):
                matcher = MATCHERS[i]
                temp = matcher(position, code)
                if to_add is None or (temp is not None and temp.end > to_add.end):
                    to_add = temp

            if to_add is not None:
                if error is not None:
                    self.tokens.append(error)
                    error = None
                self.tokens.append(to_add)
                position = to_add.end
                continue
            if error is None:
                error = Token(TokenType.ERROR, position + 1, code[position])
            else:
                error.value += code[position]
                error.end += 1
            position += 1

    # for token in self.tokens:
    #    print(token.to_string())
        return self.tokens



from .l_token import *
import re


def hex_bin_number_matcher(position, text):
    min_length = 3
    cur_pos = position
    res_position = position

    if len(text) - position < min_length:
        return None

    pattern = re.compile(r"[-+]?0[xX].?")
    match_pos = pattern.search(text[position:position + min_length], position)
    if match_pos is not None and match_pos.start() == position:
        type = TokenType.HEX
    else:
        pattern = re.compile(r"[-+]?0[bB].?")
        match_pos = pattern.search(text[position:position + min_length], position)
        if match_pos is not None and match_pos.start() == position:
            type = TokenType.BIN
        else:
            return None

    hex_pattern = re.compile(r"[-+]?0[xX][0-9a-fA-F]+([eE]([-+]?[0-9]+)?)?")
    bin_pattern = re.compile(r"[-+]?0[bB][01]+([eE]([-+]?[0-9]+)?)?")
    while cur_pos < len(text):
        cur_pos += 1
        if hex_pattern.search(text[position:cur_pos], position) is not None:
            hex_match_pos, hex_stop = hex_pattern.match(text[position:cur_pos], position)
            if type == TokenType.HEX and hex_match_pos is not None and hex_match_pos.start() == position and hex_stop == cur_pos:
                res_position = cur_pos
        '''bin_match_pos, bin_stop = bin_pattern.search(text[position:cur_pos], position)
        
        if type == TokenType.BIN and bin_match_pos is not None and hex_match_pos.start() == position:
            res_position = cur_pos'''
    return Token(type, res_position, text[position:res_position]) if res_position != position else None




import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="ccfkotlin",
    packages=["ccfkotlin"],
    entry_points={
        "console_scripts": ['ccfkotlin = ccfkotlin.__main__:main']
    },
    version="1.0.3",
    description="Kotlin code conversions fixer",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Alexandra Kmet",
    url="https://github.com/alexandrakmet/Metaprogramming",
)

